"""Functions related to path and os name setting."""
import getpass
import logging
import os
import re
import socket
from typing import Tuple

ROOT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))


def prepare_host(create_new=True, data_path=None, sampling="daily") -> str:
    """
    Set up host path.

    INFO: This functions is designed to handle known hosts. For proper working, please add your hostname hardcoded here.
    Otherwise parse your custom data_path in kwargs. If data_path is provided, hardcoded paths for known hosts will be
    ignored!

    :param create_new: Create new path if enabled
    :param data_path: Parse your custom path (and therefore ignore preset paths fitting to known hosts)
    :param sampling: sampling rate to separate data physically by temporal resolution

    :return: full path to data
    """
    hostname = get_host()
    user = getpass.getuser()
    runner_regex = re.compile(r"runner-.*-project-2411-concurrent-\d+")
    if hostname == "ZAM144":
        path = f"/home/{user}/Data/toar_{sampling}/"
    elif hostname == "zam347":
        path = f"/home/{user}/Data/toar_{sampling}/"
    elif hostname == "linux-aa9b":
        path = f"/home/{user}/machinelearningtools/data/toar_{sampling}/"
    elif (len(hostname) > 2) and (hostname[:2] == "jr"):
        path = f"/p/project/cjjsc42/{user}/DATA/toar_{sampling}/"
    elif (len(hostname) > 2) and (hostname[:2] in ['jw', 'ju'] or hostname[:5] in ['hdfml']):
        path = f"/p/project/deepacf/intelliaq/{user}/DATA/toar_{sampling}/"
    elif runner_regex.match(hostname) is not None:
        path = f"/home/{user}/machinelearningtools/data/toar_{sampling}/"
    else:
        raise OSError(f"unknown host '{hostname}'")
    if not os.path.exists(path):
        try:
            if create_new:
                check_path_and_create(path)
                return path
            else:
                raise PermissionError
        except PermissionError:
            raise NotADirectoryError(f"path '{path}' does not exist for host '{hostname}'.")
    else:
        logging.debug(f"set path to: {path}")
        return path


def set_experiment_path(name: str, path: str = None) -> str:
    """
    Set name of experiment and its path.

    * Experiment path is set to `<experiment_path>/<exp_name>` if provided or `ROOT_PATH/<exp_name>` otherwise

    :param name: custom experiment name
    :param path: custom experiment path

    :return: full experiment path
    """
    if path is None:
        experiment_path = os.path.abspath(os.path.join(ROOT_PATH, name))
    else:
        experiment_path = os.path.join(os.path.abspath(path), name)
    return experiment_path


def set_experiment_name(name: str = None, sampling: str = None) -> str:
    """
    Set name of experiment and its path.

    * Experiment name is set to `TestExperiment` if not provided. If a name is given, this string is expanded
        by suffix `_network`. Experiment name is always expanded by `_<sampling>` as ending suffix if sampling is given.

    :param name: custom experiment name
    :param sampling: sampling rate as string to add to experiment name

    :return: experiment name
    """
    if name is None:
        experiment_name = "TestExperiment"
    else:
        experiment_name = f"{name}_network"
    if sampling is not None:
        experiment_name += f"_{sampling}"
    return experiment_name


def set_bootstrap_path(bootstrap_path: str, data_path: str, sampling: str) -> str:
    """
    Set path for bootstrap input data.

    Either use given bootstrap_path or create additional folder in same directory like data path.

    :param bootstrap_path: custom path to store bootstrap data
    :param data_path: path of data for default bootstrap path
    :param sampling: sampling rate to add, if path is set to default

    :return: full bootstrap path
    """
    if bootstrap_path is None:
        bootstrap_path = os.path.join(data_path, "..", f"bootstrap_{sampling}")
    check_path_and_create(bootstrap_path)
    return os.path.abspath(bootstrap_path)


def check_path_and_create(path: str) -> None:
    """
    Check a given path and create if not existing.

    :param path: path to check and create
    """
    try:
        os.makedirs(path)
        logging.debug(f"Created path: {path}")
    except FileExistsError:
        logging.debug(f"Path already exists: {path}")


def get_host():
    return socket.gethostname()