"""Collection of stastical methods: Transformation and Skill Scores."""

from scipy import stats

__author__ = 'Lukas Leufen, Felix Kleinert'
__date__ = '2019-10-23'

import numpy as np
import xarray as xr
import pandas as pd
from typing import Union, Tuple, Dict

Data = Union[xr.DataArray, pd.DataFrame]


def apply_inverse_transformation(data: Data, mean: Data, std: Data = None, method: str = "standardise") -> Data:
    """
    Apply inverse transformation for given statistics.

    :param data: transform this data back
    :param mean: mean of transformation
    :param std: standard deviation of transformation (optional)
    :param method: transformation method

    :return: inverse transformed data
    """
    if method == 'standardise':  # pragma: no branch
        return standardise_inverse(data, mean, std)
    elif method == 'centre':  # pragma: no branch
        return centre_inverse(data, mean)
    elif method == 'normalise':  # pragma: no cover
        # use min/max of data or given min/max
        raise NotImplementedError
    else:
        raise NotImplementedError


def standardise(data: Data, dim: Union[str, int]) -> Tuple[Data, Data, Data]:
    """
    Standardise a xarray.dataarray (along dim) or pandas.DataFrame (along axis) with mean=0 and std=1.

    :param data: data to standardise
    :param dim: name (xarray) or axis (pandas) of dimension which should be standardised
    :return: mean, standard deviation and standardised data
    """
    return data.mean(dim), data.std(dim), (data - data.mean(dim)) / data.std(dim)


def standardise_inverse(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply inverse function of `standardise` on data and therefore vanishes the standardising.

    :param data: standardised data
    :param mean: mean of standardisation
    :param std: standard deviation of transformation

    :return: inverse standardised data
    """
    return data * std + mean


def standardise_apply(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply `standardise` on data using given mean and std.

    :param data: data to transform
    :param mean: mean to use for transformation
    :param std: standard deviation for transformation

    :return: transformed data
    """
    return (data - mean) / std


def centre(data: Data, dim: Union[str, int]) -> Tuple[Data, None, Data]:
    """
    Centre a xarray.dataarray (along dim) or pandas.DataFrame (along axis) to mean=0.

    :param data: data to centre
    :param dim: name (xarray) or axis (pandas) of dimension which should be centred

    :return: mean, None placeholder and centred data
    """
    return data.mean(dim), None, data - data.mean(dim)


def centre_inverse(data: Data, mean: Data) -> Data:
    """
    Apply inverse function of `centre` and therefore add given values of mean to data.

    :param data: data to apply inverse centering
    :param mean: mean to use for inverse transformation

    :return: inverted centering transformation data
    """
    return data + mean


def centre_apply(data: Data, mean: Data) -> Data:
    """
    Apply `centre` on data using given mean.

    :param data: data to transform
    :param mean: mean to use for transformation

    :return: transformed data
    """
    return data - mean


def mean_squared_error(a, b):
    """Calculate mean squared error."""
    return np.square(a - b).mean()


class SkillScores:
    r"""
    Calculate different kinds of skill scores.

    Skill score on MSE:
        Calculate skill score based on MSE for given forecast, reference and observations.

        .. math::

            \text{SkillScore} = 1 - \frac{\text{MSE(obs, for)}}{\text{MSE(obs, ref)}}

        To run:

        .. code-block:: python

            skill_scores = SkillScores(None).general_skill_score(data, observation_name, forecast_name, reference_name)

    Competitive skill score:
        Calculate skill scores to highlight differences between forecasts. This skill score is also based on the MSE.
        Currently required forecasts are CNN, OLS and persi, as well as the observation obs.

        .. code-block:: python

            skill_scores_class = SkillScores(internal_data)  # must contain columns CNN, OLS, persi and obs.
            skill_scores = skill_scores_class.skill_scores(window_lead_time=3)

    Skill score according to Murphy:
        Follow climatological skill score definition of Murphy (1988). External data is data from another time period
        than the internal data set on initialisation. In other terms, this should be the train and validation data
        whereas the internal data is the test data. This sounds perhaps counter-intuitive, but if a skill score is
        evaluated to a model to another, this must be performend test data set. Therefore, for this case the foreign
        data is train and val data.

        .. code-block:: python

            skill_scores_class = SkillScores(internal_data)  # must contain columns obs and CNN.
            skill_scores_clim = skill_scores_class.climatological_skill_scores(external_data, window_lead_time=3)

    """

    def __init__(self, internal_data: Data):
        """Set internal data."""
        self.internal_data = internal_data

    def skill_scores(self, window_lead_time: int) -> pd.DataFrame:
        """
        Calculate skill scores for all combinations of CNN, persistence and OLS.

        :param window_lead_time: length of forecast steps

        :return: skill score for each comparison and forecast step
        """
        ahead_names = list(range(1, window_lead_time + 1))
        skill_score = pd.DataFrame(index=['cnn-persi', 'ols-persi', 'cnn-ols'])
        for iahead in ahead_names:
            data = self.internal_data.sel(ahead=iahead)
            skill_score[iahead] = [self.general_skill_score(data, forecast_name="CNN", reference_name="persi"),
                                   self.general_skill_score(data, forecast_name="OLS", reference_name="persi"),
                                   self.general_skill_score(data, forecast_name="CNN", reference_name="OLS")]
        return skill_score

    def climatological_skill_scores(self, external_data: Data, window_lead_time: int) -> xr.DataArray:
        """
        Calculate climatological skill scores according to Murphy (1988).

        Calculate all CASES I - IV and terms [ABC][I-IV]. Internal data has to be set by initialisation, external data
        is part of parameters.

        :param external_data: external data
        :param window_lead_time: interested time step of forecast horizon to select data

        :return: all CASES as well as all terms
        """
        ahead_names = list(range(1, window_lead_time + 1))

        all_terms = ['AI', 'AII', 'AIII', 'AIV', 'BI', 'BII', 'BIV', 'CI', 'CIV', 'CASE I', 'CASE II', 'CASE III',
                     'CASE IV']
        skill_score = xr.DataArray(np.full((len(all_terms), len(ahead_names)), np.nan), coords=[all_terms, ahead_names],
                                   dims=['terms', 'ahead'])

        for iahead in ahead_names:

            data = self.internal_data.sel(ahead=iahead)

            skill_score.loc[["CASE I", "AI", "BI", "CI"], iahead] = np.stack(self._climatological_skill_score(
                data, mu_type=1, forecast_name="CNN").values.flatten())

            skill_score.loc[["CASE II", "AII", "BII"], iahead] = np.stack(self._climatological_skill_score(
                data, mu_type=2, forecast_name="CNN").values.flatten())

            if external_data is not None:
                skill_score.loc[["CASE III", "AIII"], iahead] = np.stack(self._climatological_skill_score(
                    data, mu_type=3, forecast_name="CNN",
                    external_data=external_data).values.flatten())

                skill_score.loc[["CASE IV", "AIV", "BIV", "CIV"], iahead] = np.stack(self._climatological_skill_score(
                    data, mu_type=4, forecast_name="CNN",
                    external_data=external_data).values.flatten())

        return skill_score

    def _climatological_skill_score(self, data, mu_type=1, observation_name="obs", forecast_name="CNN",
                                    external_data=None):
        kwargs = {"external_data": external_data} if external_data is not None else {}
        return self.__getattribute__(f"skill_score_mu_case_{mu_type}")(data, observation_name, forecast_name, **kwargs)

    @staticmethod
    def general_skill_score(data: Data, observation_name: str = "obs", forecast_name: str = "CNN",
                            reference_name: str = "persi") -> np.ndarray:
        r"""
        Calculate general skill score based on mean squared error.

        :param data: internal data containing data for observation, forecast and reference
        :param observation_name: name of observation
        :param forecast_name: name of forecast
        :param reference_name: name of reference

        :return: skill score of forecast
        """
        data = data.dropna("index")
        observation = data.sel(type=observation_name)
        forecast = data.sel(type=forecast_name)
        reference = data.sel(type=reference_name)
        mse = mean_squared_error
        skill_score = 1 - mse(observation, forecast) / mse(observation, reference)
        return skill_score.values

    @staticmethod
    def skill_score_pre_calculations(data: Data, observation_name: str, forecast_name: str) -> Tuple[np.ndarray,
                                                                                                     np.ndarray,
                                                                                                     np.ndarray,
                                                                                                     Data,
                                                                                                     Dict[str, Data]]:
        """
        Calculate terms AI, BI, and CI, mean, variance and pearson's correlation and clean up data.

        The additional information on mean, variance and pearson's correlation (and the p-value) are returned as
        dictionary with the corresponding keys mean, sigma, r and p.

        :param data: internal data to use for calculations
        :param observation_name: name of observation
        :param forecast_name: name of forecast

        :returns: Terms AI, BI, and CI, internal data without nans and mean, variance, correlation and its p-value
        """
        data = data.loc[..., [observation_name, forecast_name]].drop("ahead")
        data = data.dropna("index")

        mean = data.mean("index")
        sigma = np.sqrt(data.var("index"))
        r, p = stats.pearsonr(data.loc[..., forecast_name], data.loc[..., observation_name])

        AI = np.array(r ** 2)
        BI = ((r - (sigma.loc[..., forecast_name] / sigma.loc[..., observation_name])) ** 2).values
        CI = (((mean.loc[..., forecast_name] - mean.loc[..., observation_name]) / sigma.loc[
            ..., observation_name]) ** 2).values

        suffix = {"mean": mean, "sigma": sigma, "r": r, "p": p}
        return AI, BI, CI, data, suffix

    def skill_score_mu_case_1(self, data, observation_name="obs", forecast_name="CNN"):
        """Calculate CASE I."""
        AI, BI, CI, data, _ = self.skill_score_pre_calculations(data, observation_name, forecast_name)
        skill_score = np.array(AI - BI - CI)
        return pd.DataFrame({"skill_score": [skill_score], "AI": [AI], "BI": [BI], "CI": [CI]}).to_xarray().to_array()

    def skill_score_mu_case_2(self, data, observation_name="obs", forecast_name="CNN"):
        """Calculate CASE II."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(data, observation_name, forecast_name)
        monthly_mean = self.create_monthly_mean_from_daily_data(data)
        data = xr.concat([data, monthly_mean], dim="type")
        sigma = suffix["sigma"]
        sigma_monthly = np.sqrt(monthly_mean.var())
        r, p = stats.pearsonr(data.loc[..., observation_name], data.loc[..., observation_name + "X"])
        AII = np.array(r ** 2)
        BII = ((r - sigma_monthly / sigma.loc[observation_name]) ** 2).values
        skill_score = np.array((AI - BI - CI - AII + BII) / (1 - AII + BII))
        return pd.DataFrame({"skill_score": [skill_score], "AII": [AII], "BII": [BII]}).to_xarray().to_array()

    def skill_score_mu_case_3(self, data, observation_name="obs", forecast_name="CNN", external_data=None):
        """Calculate CASE III."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(data, observation_name, forecast_name)
        mean, sigma = suffix["mean"], suffix["sigma"]
        AIII = (((external_data.mean().values - mean.loc[observation_name]) / sigma.loc[observation_name]) ** 2).values
        skill_score = np.array((AI - BI - CI + AIII) / 1 + AIII)
        return pd.DataFrame({"skill_score": [skill_score], "AIII": [AIII]}).to_xarray().to_array()

    def skill_score_mu_case_4(self, data, observation_name="obs", forecast_name="CNN", external_data=None):
        """Calculate CASE IV."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(data, observation_name, forecast_name)
        monthly_mean_external = self.create_monthly_mean_from_daily_data(external_data, columns=data.type.values,
                                                                         index=data.index)
        data = xr.concat([data, monthly_mean_external], dim="type")
        mean, sigma = suffix["mean"], suffix["sigma"]
        monthly_mean_external = self.create_monthly_mean_from_daily_data(external_data, columns=data.type.values)
        mean_external = monthly_mean_external.mean()
        sigma_external = np.sqrt(monthly_mean_external.var())

        # r_mu, p_mu = stats.spearmanr(data.loc[..., [observation_name, observation_name+'X']])
        r_mu, p_mu = stats.pearsonr(data.loc[..., observation_name], data.loc[..., observation_name + "X"])

        AIV = np.array(r_mu ** 2)
        BIV = ((r_mu - sigma_external / sigma.loc[observation_name]) ** 2).values
        CIV = (((mean_external - mean.loc[observation_name]) / sigma.loc[observation_name]) ** 2).values
        skill_score = np.array((AI - BI - CI - AIV + BIV + CIV) / (1 - AIV + BIV + CIV))
        return pd.DataFrame(
            {"skill_score": [skill_score], "AIV": [AIV], "BIV": [BIV], "CIV": CIV}).to_xarray().to_array()

    @staticmethod
    def create_monthly_mean_from_daily_data(data, columns=None, index=None):
        """
        Calculate average for each month and save as daily values with flag 'X'.

        :param data: data to average
        :param columns: columns to work on (all columns from given data are used if empty)
        :param index: index of returned data (index of given data is used if empty)

        :return: data containing monthly means in daily resolution
        """
        if columns is None:
            columns = data.type.values
        if index is None:
            index = data.index
        coordinates = [index, [v + "X" for v in list(columns)]]
        empty_data = np.full((len(index), len(columns)), np.nan)
        monthly_mean = xr.DataArray(empty_data, coords=coordinates, dims=["index", "type"])
        mu = data.groupby("index.month").mean()

        for month in mu.month:
            monthly_mean[monthly_mean.index.dt.month == month, :] = mu[mu.month == month].values

        return monthly_mean
