import operator as op
import os
import pickle

import numpy as np
import pytest
import xarray as xr

from src.data_handling.data_generator import DataGenerator
from src.data_handling.data_preparation import DataPrep
from src.helpers.join import EmptyQueryResult


class TestDataGenerator:

    # @pytest.fixture(autouse=True, scope='module')
    # def teardown_module(self):
    #     yield
    #     if "data" in os.listdir(os.path.dirname(__file__)):
    #         shutil.rmtree(os.path.join(os.path.dirname(__file__), "data"), ignore_errors=True)

    @pytest.fixture
    def gen(self):
        return DataGenerator(os.path.join(os.path.dirname(__file__), 'data'), 'AIRBASE', 'DEBW107', ['o3', 'temp'],
                             'datetime', 'variables', 'o3', start=2010, end=2014)

    @pytest.fixture
    def gen_with_transformation(self):
        return DataGenerator(os.path.join(os.path.dirname(__file__), 'data'), 'AIRBASE', 'DEBW107', ['o3', 'temp'],
                             'datetime', 'variables', 'o3', start=2010, end=2014,
                             transformation={"scope": "data", "mean": "estimate"},
                             statistics_per_var={'o3': 'dma8eu', 'temp': 'maximum'})

    @pytest.fixture
    def gen_no_init(self):
        generator = object.__new__(DataGenerator)
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))
        generator.data_path = path
        if not os.path.exists(path):
            os.makedirs(path)
        generator.stations = ["DEBW107", "DEBW013", "DEBW001"]
        generator.network = "AIRBASE"
        generator.variables = ["temp", "o3"]
        generator.station_type = "background"
        generator.kwargs = {"start": 2010, "end": 2014, "statistics_per_var": {'o3': 'dma8eu', 'temp': 'maximum'}}
        return generator

    @pytest.fixture
    def accurate_transformation(self, gen_no_init):
        tmp = np.nan
        for station in gen_no_init.stations:
            try:
                data_prep = DataPrep(gen_no_init.data_path, gen_no_init.network, station, gen_no_init.variables,
                                     station_type=gen_no_init.station_type, **gen_no_init.kwargs)
                tmp = data_prep.data.combine_first(tmp)
            except EmptyQueryResult:
                continue
        mean_expected = tmp.mean(dim=["Stations", "datetime"])
        std_expected = tmp.std(dim=["Stations", "datetime"])
        return mean_expected, std_expected

    @pytest.fixture
    def estimated_transformation(self, gen_no_init):
        mean, std = None, None
        for station in gen_no_init.stations:
            try:
                data_prep = DataPrep(gen_no_init.data_path, gen_no_init.network, station, gen_no_init.variables,
                                     station_type=gen_no_init.station_type, **gen_no_init.kwargs)
                mean = data_prep.data.mean(axis=1).combine_first(mean)
                std = data_prep.data.std(axis=1).combine_first(std)
            except EmptyQueryResult:
                continue
        mean_expected = mean.mean(axis=0)
        std_expected = std.mean(axis=0)
        return mean_expected, std_expected

    class DummyDataPrep:
        def __init__(self, data):
            self.station = "DEBW107"
            self.variables = ["o3", "temp"]
            self.data = data

    def test_init(self, gen):
        assert gen.data_path == os.path.join(os.path.dirname(__file__), 'data')
        assert gen.network == 'AIRBASE'
        assert gen.stations == ['DEBW107']
        assert gen.variables == ['o3', 'temp']
        assert gen.station_type is None
        assert gen.interpolate_dim == 'datetime'
        assert gen.target_dim == 'variables'
        assert gen.target_var == 'o3'
        assert gen.interpolate_method == "linear"
        assert gen.limit_nan_fill == 1
        assert gen.window_history_size == 7
        assert gen.window_lead_time == 4
        assert gen.transformation is None
        assert gen.kwargs == {"start": 2010, "end": 2014}

    def test_repr(self, gen):
        path = os.path.join(os.path.dirname(__file__), 'data')
        assert gen.__repr__().rstrip() == f"DataGenerator(path='{path}', network='AIRBASE', stations=['DEBW107'], " \
                                          f"variables=['o3', 'temp'], station_type=None, interpolate_dim='datetime', " \
                                          f"target_dim='variables', target_var='o3', **{{'start': 2010, 'end': 2014}})" \
            .rstrip()

    def test_len(self, gen):
        assert len(gen) == 1
        gen.stations = ['station1', 'station2', 'station3']
        assert len(gen) == 3

    def test_iter(self, gen):
        assert hasattr(gen, '_iterator') is False
        iter(gen)
        assert hasattr(gen, '_iterator')
        assert gen._iterator == 0

    def test_next(self, gen):
        gen.kwargs = {'statistics_per_var': {'o3': 'dma8eu', 'temp': 'maximum'}}
        for i, d in enumerate(gen, start=1):
            assert i == gen._iterator

    def test_getitem(self, gen):
        gen.kwargs = {'statistics_per_var': {'o3': 'dma8eu', 'temp': 'maximum'}}
        station = gen["DEBW107"]
        assert len(station) == 2
        assert station[0].Stations.data == "DEBW107"
        assert station[0].data.shape[1:] == (8, 1, 2)
        assert station[1].data.shape[-1] == gen.window_lead_time
        assert station[0].data.shape[1] == gen.window_history_size + 1

    def test_setup_transformation_no_transformation(self, gen_no_init):
        assert gen_no_init.setup_transformation(None) is None
        assert gen_no_init.setup_transformation({}) == {"method": "standardise", "mean": None, "std": None}
        assert gen_no_init.setup_transformation({"scope": "station", "mean": "accurate"}) == \
               {"scope": "station", "method": "standardise", "mean": None, "std": None}

    def test_setup_transformation_calculate_statistics(self, gen_no_init):
        transformation = {"scope": "data", "mean": "accurate"}
        res_acc = gen_no_init.setup_transformation(transformation)
        assert sorted(res_acc.keys()) == sorted(["scope", "mean", "std", "method"])
        assert isinstance(res_acc["mean"], xr.DataArray)
        assert isinstance(res_acc["std"], xr.DataArray)
        transformation["mean"] = "estimate"
        res_est = gen_no_init.setup_transformation(transformation)
        assert sorted(res_est.keys()) == sorted(["scope", "mean", "std", "method"])
        assert isinstance(res_est["mean"], xr.DataArray)
        assert isinstance(res_est["std"], xr.DataArray)
        assert np.testing.assert_array_compare(op.__ne__, res_est["std"].values, res_acc["std"].values) is None

    def test_setup_transformation_use_given_statistics(self, gen_no_init):
        mean = xr.DataArray([30, 15], coords={"variables": ["o3", "temp"]}, dims=["variables"])
        transformation = {"scope": "data", "method": "centre", "mean": mean}
        res = gen_no_init.setup_transformation(transformation)
        assert np.testing.assert_equal(res["mean"].values, mean.values) is None
        assert res["std"] is None

    def test_setup_transformation_errors(self, gen_no_init):
        transformation = {"scope": "random", "mean": "accurate"}
        with pytest.raises(ValueError):
            gen_no_init.setup_transformation(transformation)
        transformation = {"scope": "data", "mean": "fit"}
        with pytest.raises(ValueError):
            gen_no_init.setup_transformation(transformation)

    def test_calculate_accurate_transformation_standardise(self, gen_no_init, accurate_transformation):
        mean_expected, std_expected = accurate_transformation
        mean, std = gen_no_init.calculate_accurate_transformation("standardise")
        assert np.testing.assert_almost_equal(mean_expected.values, mean.values) is None
        assert np.testing.assert_almost_equal(std_expected.values, std.values) is None

    def test_calculate_accurate_transformation_centre(self, gen_no_init, accurate_transformation):
        mean_expected, _ = accurate_transformation
        mean, std = gen_no_init.calculate_accurate_transformation("centre")
        assert np.testing.assert_almost_equal(mean_expected.values, mean.values) is None
        assert std is None

    def test_calculate_accurate_transformation_all_others(self, gen_no_init):
        with pytest.raises(NotImplementedError):
            gen_no_init.calculate_accurate_transformation("normalise")

    def test_calculate_estimated_transformation_standardise(self, gen_no_init, estimated_transformation):
        mean_expected, std_expected = estimated_transformation
        mean, std = gen_no_init.calculate_estimated_transformation("standardise")
        assert np.testing.assert_almost_equal(mean_expected.values, mean.values) is None
        assert np.testing.assert_almost_equal(std_expected.values, std.values) is None

    def test_calculate_estimated_transformation_centre(self, gen_no_init, estimated_transformation):
        mean_expected, _ = estimated_transformation
        mean, std = gen_no_init.calculate_estimated_transformation("centre")
        assert np.testing.assert_almost_equal(mean_expected.values, mean.values) is None
        assert std is None

    def test_calculate_estimated_transformation_all_others(self, gen_no_init):
        with pytest.raises(NotImplementedError):
            gen_no_init.calculate_estimated_transformation("normalise")

    def test_get_station_key(self, gen):
        gen.stations.append("DEBW108")
        f = gen.get_station_key
        iter(gen)
        assert f(None) == "DEBW107"
        with pytest.raises(KeyError) as e:
            f([None, None])
        assert "More than one key was given: [None, None]" in e.value.args[0]
        assert f(1) == "DEBW108"
        assert f([1]) == "DEBW108"
        with pytest.raises(KeyError) as e:
            f(3)
        assert "3 is not in range(0, 2)" in e.value.args[0]
        assert f("DEBW107") == "DEBW107"
        assert f(["DEBW108"]) == "DEBW108"
        with pytest.raises(KeyError) as e:
            f("DEBW999")
        assert "DEBW999 is not in stations" in e.value.args[0]
        with pytest.raises(KeyError) as e:
            f(6.5)
        assert "key has to be from Union[str, int]. Given was 6.5 (float)"

    def test_get_data_generator(self, gen):
        gen.kwargs = {"statistics_per_var": {'o3': 'dma8eu', 'temp': 'maximum'}}
        file = os.path.join(gen.data_path_tmp, f"DEBW107_{'_'.join(sorted(gen.variables))}_None_None_.pickle")
        if os.path.exists(file):
            os.remove(file)
        assert not os.path.exists(file)
        assert isinstance(gen.get_data_generator("DEBW107", load_local_tmp_storage=False), DataPrep)
        t = os.stat(file).st_ctime
        assert os.path.exists(file)
        assert isinstance(gen.get_data_generator("DEBW107"), DataPrep)
        assert os.stat(file).st_mtime == t
        os.remove(file)
        assert isinstance(gen.get_data_generator("DEBW107"), DataPrep)
        assert os.stat(file).st_ctime > t

    def test_get_data_generator_transform(self, gen_with_transformation):
        gen = gen_with_transformation
        data = gen.get_data_generator("DEBW107", load_local_tmp_storage=False, save_local_tmp_storage=False)
        assert data._transform_method == "standardise"
        assert data.mean is not None

    def test_get_data_generator_extremes(self, gen_with_transformation):
        gen = gen_with_transformation
        gen.kwargs = {"statistics_per_var": {'o3': 'dma8eu', 'temp': 'maximum'}}
        gen.extreme_values = [1.]
        data = gen.get_data_generator("DEBW107", load_local_tmp_storage=False, save_local_tmp_storage=False)
        assert data.extremes_label is not None
        assert data.extremes_history is not None
        assert data.extremes_label.shape[:2] == data.label.shape[:2]
        assert data.extremes_label.shape[2] <= data.label.shape[2]
        len_both_tails = data.extremes_label.shape[2]
        gen.kwargs["extremes_on_right_tail_only"] = True
        data = gen.get_data_generator("DEBW107", load_local_tmp_storage=False, save_local_tmp_storage=False)
        assert data.extremes_label.shape[2] <= len_both_tails

    def test_save_pickle_data(self, gen):
        file = os.path.join(gen.data_path_tmp, f"DEBW107_{'_'.join(sorted(gen.variables))}_2010_2014_.pickle")
        if os.path.exists(file):
            os.remove(file)
        assert not os.path.exists(file)
        data = self.DummyDataPrep(np.ones((10, 2)))
        gen._save_pickle_data(data)
        assert os.path.exists(file)
        os.remove(file)

    def test_load_pickle_data(self, gen):
        file = os.path.join(gen.data_path_tmp, f"DEBW107_{'_'.join(sorted(gen.variables))}_2010_2014_.pickle")
        data = self.DummyDataPrep(np.ones((10, 2)))
        with open(file, "wb") as f:
            pickle.dump(data, f)
        assert os.path.exists(file)
        res = gen._load_pickle_data("DEBW107", ["o3", "temp"]).data
        assert np.testing.assert_almost_equal(res, np.ones((10, 2))) is None
        os.remove(file)
