import numpy as np
import pandas as pd
import pytest
import xarray as xr

from src.helpers.statistics import standardise, standardise_inverse, standardise_apply, centre, centre_inverse, centre_apply, \
    apply_inverse_transformation

lazy = pytest.lazy_fixture


@pytest.fixture(scope='module')
def input_data():
    return np.array([np.random.normal(2, 2, 3000),
                     np.random.normal(-5, 3, 3000),
                     np.random.normal(10, 1, 3000)]).T


@pytest.fixture(scope='module')
def pandas(input_data):
    return pd.DataFrame(input_data)


@pytest.fixture(scope='module')
def pd_mean():
    return [2, 10, 3]


@pytest.fixture(scope='module')
def pd_std():
    return [3, 2, 3]


@pytest.fixture(scope='module')
def xarray(input_data):
    shape = input_data.shape
    coords = {'index': range(shape[0]), 'value': range(shape[1])}
    return xr.DataArray(input_data, coords=coords, dims=coords.keys())


@pytest.fixture(scope='module')
def xr_mean(input_data):
    return xr.DataArray([2, 10, 3], coords={'value': range(3)}, dims=['value'])


@pytest.fixture(scope='module')
def xr_std(input_data):
    return xr.DataArray([3, 2, 3], coords={'value': range(3)}, dims=['value'])


class TestStandardise:

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_standardise(self, data_orig, dim):
        mean, std, data = standardise(data_orig, dim)
        assert np.testing.assert_almost_equal(mean, [2, -5, 10], decimal=1) is None
        assert np.testing.assert_almost_equal(std, [2, 3, 1], decimal=1) is None
        assert np.testing.assert_almost_equal(data.mean(dim), [0, 0, 0]) is None
        assert np.testing.assert_almost_equal(data.std(dim), [1, 1, 1]) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_standardise_inverse(self, data_orig, dim):
        mean, std, data = standardise(data_orig, dim)
        data_recovered = standardise_inverse(data, mean, std)
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_apply_standardise_inverse(self, data_orig, dim):
        mean, std, data = standardise(data_orig, dim)
        data_recovered = apply_inverse_transformation(data, mean, std)
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, mean, std, dim', [(lazy('pandas'), lazy('pd_mean'), lazy('pd_std'), 0),
                                                           (lazy('xarray'), lazy('xr_mean'), lazy('xr_std'), 'index')])
    def test_standardise_apply(self, data_orig, mean, std, dim):
        data = standardise_apply(data_orig, mean, std)
        mean_expected = (np.array([2, -5, 10]) - np.array([2, 10, 3])) / np.array([3, 2, 3])
        std_expected = np.array([2, 3, 1]) / np.array([3, 2, 3])
        assert np.testing.assert_almost_equal(data.mean(dim), mean_expected, decimal=1) is None
        assert np.testing.assert_almost_equal(data.std(dim), std_expected, decimal=1) is None


class TestCentre:

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_centre(self, data_orig, dim):
        mean, std, data = centre(data_orig, dim)
        assert np.testing.assert_almost_equal(mean, [2, -5, 10], decimal=1) is None
        assert std is None
        assert np.testing.assert_almost_equal(data.mean(dim), [0, 0, 0]) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_centre_inverse(self, data_orig, dim):
        mean, _, data = centre(data_orig, dim)
        data_recovered = centre_inverse(data, mean)
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, dim', [(lazy('pandas'), 0),
                                                (lazy('xarray'), 'index')])
    def test_apply_centre_inverse(self, data_orig, dim):
        mean, _, data = centre(data_orig, dim)
        data_recovered = apply_inverse_transformation(data, mean, method="centre")
        assert np.testing.assert_array_almost_equal(data_orig, data_recovered) is None

    @pytest.mark.parametrize('data_orig, mean, dim', [(lazy('pandas'), lazy('pd_mean'), 0),
                                                      (lazy('xarray'), lazy('xr_mean'), 'index')])
    def test_centre_apply(self, data_orig, mean, dim):
        data = centre_apply(data_orig, mean)
        mean_expected = np.array([2, -5, 10]) - np.array([2, 10, 3])
        assert np.testing.assert_almost_equal(data.mean(dim), mean_expected, decimal=1) is None
